import 'dart:async';
import 'dart:convert';
import 'dart:math';

// ignore: avoid_web_libraries_in_flutter
import 'dart:html' as html;
import 'package:convert/convert.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:pointycastle/export.dart';
import 'package:pointycastle/export.dart' as pc;
import 'package:robomotion_app/encryption.dart';
import 'package:robomotion_app/errors.dart';
import 'package:robomotion_app/exception.dart';
import 'package:robomotion_app/models/instance_upload_link.dart';
import 'package:robomotion_app/srp/client.dart';
import 'package:robomotion_app/srp/standard_groups.dart';
import 'package:robomotion_app/utils/extensions.dart';
import 'package:version/version.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import 'package:uuid/uuid.dart';
import 'package:encrypt/encrypt.dart' as enc;
import 'package:fast_rsa/fast_rsa.dart' as rsa;
import 'package:cryptography/cryptography.dart';
import 'package:bs58/bs58.dart';
import 'package:hex/hex.dart';
import 'models/published_flow.dart';
import 'models/get_instance.dart';
import 'models/start_instance.dart';
import 'models/queue_message.dart';

const version = "0.8.5";

typedef Future<Uint8List> EncryptionCallback(Uint8List data, Uint8List key);
typedef Future<Uint8List> DecryptionCallback(Uint8List data, Uint8List key);

StreamController<void> _keyExStreamCtrl = StreamController<void>.broadcast();
StreamController<void> _robotIDStreamCtrl = StreamController<void>.broadcast();

class Vault {
  final Map vault;
  List<int>? secretKey;
  List<int>? _vaultKey;
  String encSecretKey;
  String encVaultKey;

  Vault({
    required this.vault,
    this.secretKey,
    List<int>? vaultKey,
    this.encSecretKey = '',
    this.encVaultKey = '',
  }) : _vaultKey = vaultKey;

  Future<List<int>> getVaultKey(Map user, String password) async {
    if (_vaultKey != null) {
      return _vaultKey!;
    }

    encVaultKey = vault['enc_vault_key'];
    if (encVaultKey.isEmpty) {
      throw Exception("Empty encrypted vault key.");
    }

    String pkPem = await Robomotion().getPrivateKey(
      user,
      password,
    );

    final vk = await rsa.RSA.decryptOAEPBytes(
      base64Decode(encVaultKey),
      "",
      rsa.Hash.SHA256,
      pkPem,
    );

    _vaultKey = vk.toList();
    return _vaultKey!;
  }

  Future<List<int>> getSecretKey(Map user, String password) async {
    if (this.secretKey != null) {
      return this.secretKey!;
    }

    if (this.encSecretKey.isEmpty) {
      throw Exception("Empty encrypted vault secret key.");
    }

    String pkPem = await Robomotion().getPrivateKey(
      user,
      password,
    );

    final sk = await rsa.RSA.decryptOAEPBytes(
      base64Decode(this.encSecretKey),
      "",
      rsa.Hash.SHA256,
      pkPem,
    );

    this.secretKey = sk.toList();
    return this.secretKey!;
  }
}

class Robomotion {
  Config? config;

  void Function(QueueMessage data)? onMsgRecv;

  String? get _apiURL {
    return _getAPIUrl();
  }

  String? get _appURL {
    return _getAppUrl();
  }

  String? get _sid {
    if (config!.token != null && config!.token!.isNotEmpty) {
      return config!.token;
    } else {
      return _getCookie("sid");
    }
  }

  late WebSocketChannel _channel;
  final Map<String, StreamController<dynamic>> _streams = {};
  late EncryptionCallback? encrypt;
  late DecryptionCallback? decrypt;

  Robomotion({this.config, this.onMsgRecv, this.encrypt, this.decrypt}) {
    config ??= Config();
    generateAESKey();

    this.encrypt ??= (data, key) => Future(() => encryptPayload(data));
    this.decrypt ??= (data, key) => Future(() => decryptPayload(data));

    this.onMsgRecv ??= (msg) {
      if (_streams.containsKey(msg.id)) {
        var payload = jsonDecode(utf8.decode(msg.payload!));
        _streams[msg.id]?.add(payload);
      }
    };
  }

  String? getInstanceIDFromQueryParam() {
    try {
      String currentUrl = html.window.location.href;
      Uri uri = Uri.parse(currentUrl);
      String instanceID = uri.query;
      String base58Address = instanceID;

      final Uint8List decodedBase58 = base58.decode(base58Address);
      return HEX.encode(decodedBase58);
    } catch (err) {
      print(err.toString());
      return null;
    }
  }

  Future<GetInstanceResponse> getAppInstance(String instanceID) async {
    String url = "${this._apiURL}/v1/apps.instance.get";
    Map resp = await _request(
      method: "GET",
      url: url,
      queryParams: {"id": instanceID},
      throwException: false,
    );

    return GetInstanceResponse.fromJson(resp);
  }

  Future<StartInstanceResponse> startInstance(String instanceID) async {
    String url = "${this._apiURL}/v1/apps.instance.start";
    Map resp = await _request(
      method: "POST",
      url: url,
      body: {
        "id": instanceID,
      },
      throwException: false,
    );

    return StartInstanceResponse.fromJson(resp);
  }

  Future<void> connect(
      {String? token,
      String? instanceID,
      String? apiURL = "https://api.robomotion.io"}) async {
    config = Config(workspace: apiURL, token: token);

    String? queriedInstanceID = getInstanceIDFromQueryParam();

    try {
      if (instanceID != null &&
          (queriedInstanceID == null || queriedInstanceID.isEmpty)) {
        final Uint8List decodedBase58 = base58.decode(instanceID);
        config!.instanceID = HEX.encode(decodedBase58);
      } else {
        config!.instanceID = queriedInstanceID;
      }
    } catch (err) {}

    if (config!.instanceID == null || config!.instanceID!.isEmpty) {
      throw EmptyInstanceID();
    }

    GetInstanceResponse respGetInstance =
        await getAppInstance(config!.instanceID!);
    if (respGetInstance.ok == false) {
      throw ApiResponseException(respGetInstance.error!);
    }
    if (respGetInstance.data != null) {
      config!.appID = respGetInstance.data!.appId;
      config!.robotID = respGetInstance.data!.robotId;
    }

    StartInstanceResponse respStartInstance =
        await startInstance(config!.instanceID!);
    if (respStartInstance.runId == null && respStartInstance.ok == false) {
      throw ApiResponseException(respStartInstance.error!);
    }
    config!.runID = respStartInstance.runId;

    await this.run(config!.appID!);
    await this.keyExchange();

    return;
  }

  Future<void> run(String appId) async {
    config!.appID = appId;
    await _connectAppServer(appId);
  }

  Future<Map> login(String domain, String username, String password) async {
    String nonce = Uuid().v4().replaceAll("-", "") +
        (DateTime.now().millisecondsSinceEpoch / 1000).toString();
    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/auth.challenge',
      body: {
        "username": username,
        "domain": domain,
        "nonce": nonce,
      },
    );

    var client = SRP6Client(
      group: SRP6StandardGroups.rfc5054_4096,
      digest: SHA256Digest(),
      random: _getSecureRandom(),
    );

    var A = await client.generateClientCredentialsAsync(
      Uint8List.fromList(hex.decode(resp["salt"])),
      Uint8List.fromList(utf8.encode(username)),
      Uint8List.fromList(utf8.encode(password)),
    );
    A = A ?? BigInt.zero;

    client.calculateSecret(BigInt.parse(resp["b"], radix: 16));
    var m1 = client.calculateClientEvidenceMessage() ?? BigInt.zero;

    var loginResp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/auth.login',
      body: {
        "username": username,
        "domain": domain,
        "nonce": nonce,
        "M1": m1.toRadixString(16),
        "A": A.toRadixString(16),
      },
    );

    return loginResp;
  }

  pc.SecureRandom _getSecureRandom() {
    var secureRandom = FortunaRandom();
    var random = Random.secure();
    List<int> seeds = [];
    for (int i = 0; i < 32; i++) {
      seeds.add(random.nextInt(255));
    }
    secureRandom.seed(KeyParameter(Uint8List.fromList(seeds)));
    return secureRandom;
  }

  Future<List<PublishedFlow>> getPublishedFlows() async {
    String url = "${this._apiURL}/v1/flows.published.list";
    Map resp = await _request(
      method: "GET",
      url: url,
      queryParams: {"page": "1", "size": "100", "search": ""},
    );

    List<PublishedFlow> flows = [];
    for (dynamic f in resp["flows"]) {
      flows.add(PublishedFlow.fromJson(f));
    }
    return flows;
  }

  Future<PublishedFlow> getPublishedFlow(String id) async {
    String url = "${this._apiURL}/v1/flows.published.get";
    Map resp = await _request(
      method: "GET",
      url: url,
      queryParams: {"id": id},
    );

    return PublishedFlow.fromJson(resp["flow"]);
  }

  Future<bool> publishAppInstance(String instanceID, bool published) async {
    String url = "${this._apiURL}/v1/apps.instance.publish";
    Map resp = await _request(
      method: "POST",
      url: url,
      body: {
        "instance_id": instanceID,
        "published": published,
      },
    );

    return resp["ok"] == true;
  }

  Future<String?> runFlow(
    String instanceID,
    String flowID,
    String versionID,
  ) async {
    String url = "${this._apiURL}/v1/apps.instance.start";
    String? _flowID = flowID;
    Map resp;

    try {
      resp = await _request(
        method: "POST",
        url: url,
        body: {"id": instanceID, "flow_id": flowID, "version_id": versionID},
      );
    } on AuthFailedException catch (err) {
      throw err;
    } on ApiException catch (err) {
      throw err;
    } catch (err) {
      throw err;
    }

    config!.runID = resp["run_id"];
    return _flowID;
  }

  Future<Map> apiRequest({
    required String method,
    required String endpoint,
    Map? body,
    Map<String, String>? queryParams,
    bool withCredentials = true,
  }) async {
    if (!endpoint.startsWith("/")) {
      endpoint = "/$endpoint";
    }

    return _request(
      method: method,
      url: "$_apiURL$endpoint",
      body: body,
      queryParams: queryParams,
      withCredentials: withCredentials,
    );
  }

  Future<Map> _request({
    required String method,
    required String url,
    Map<dynamic, dynamic>? body,
    Map<String, String>? queryParams,
    bool withCredentials = true,
    bool throwException = true,
  }) async {
    Map<String, String> headers = {
      'Content-Type': 'application/json; charset=UTF-8',
    };

    // If token is given as param it should be used or else lib should get sid from cookie
    if ((kDebugMode && _sid != null) ||
        (config!.token != null && config!.token!.isNotEmpty)) {
      headers['Authorization'] = 'Bearer $_sid';
    }
    Uri uri = Uri.parse(url);
    if (queryParams != null) {
      uri = Uri.parse(url).replace(queryParameters: queryParams);
    }
    var response = await html.HttpRequest.request(
      uri.toString(),
      method: method,
      withCredentials: withCredentials,
      requestHeaders: headers,
      mimeType: "application/json",
      responseType: "json",
      sendData: body != null ? json.encode(body) : null,
    );

    if (response.response == null) {
      throw new NoResponseException();
    }

    var respBody = response.response;
    if (throwException && (respBody["ok"] == null || !respBody["ok"])) {
      String errorCode = respBody["error"] ?? respBody["code"];
      if (errorCode == ErrAuthFailed) {
        throw new AuthFailedException();
      }

      throw new ApiException(errorCode, body: respBody);
    }

    return respBody;
  }

  Future<void> _connectAppServer(String appId) async {
    String url = "$_appURL?id=$appId";
    if (kDebugMode || (config!.token != null && config!.token!.isNotEmpty)) {
      url += "&sid=$_sid";
    }

    _channel = WebSocketChannel.connect(Uri.parse(url));
    _channel.stream.listen((data) async {
      final msg = QueueMessage.fromJson(jsonDecode(data));
      msg.robomotion = this;

      if (msg.encrypted!) {
        msg.payload = await this.decrypt!(msg.payload!, aesKey.bytes);
      }

      _onMsgRecv(msg).then((_) {
        if (onMsgRecv != null) onMsgRecv!(msg);
      });
    });
  }

  Future<void> _onMsgRecv(QueueMessage msg) async {
    if (msg.msgType == QMT_Error) {
      return;
    }

    if (config!.robotID == "" && msg.msgType == QMT_RobotID) {
      await _robotIDStreamCtrl.stream.first;
    }

    switch (msg.msgType) {
      case QMT_KeyExchange:
        final pem = msg.payload!;
        robotPubKeyPem = utf8.decode(pem);
        robotPubKey = parsePublicKey(pem);
        config!.version = msg.version;
        _keyExStreamCtrl.add(null);
        break;

      case QMT_RobotID:
        config!.robotID = utf8.decode(msg.payload!);
        _robotIDStreamCtrl.add(null);
        break;
    }
  }

  Future<void> sendAppMsg(QueueMessage msg) async {
    if (msg.encrypted!) {
      msg.payload = await this.encrypt!(msg.payload!, aesKey.bytes);
    }
    _channel.sink.add(msg.toString());
  }

  Future<void> keyExchange() async {
    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: config!.appID,
      robotID: config!.robotID,
      userID: config!.userID,
      msgType: QMT_KeyExchange,
      payload: null,
      encrypted: false,
      version: config?.version,
    );

    await sendAppMsg(msg);
    await _keyExStreamCtrl.stream.first;

    final encAes = rsaEncrypt(robotPubKey, aesKey.bytes);
    msg = QueueMessage(
      id: Uuid().v4(),
      appID: config!.appID,
      robotID: config!.robotID,
      msgType: QMT_KeyExchange,
      payload: encAes,
      encrypted: false,
      version: config?.version,
    );

    await sendAppMsg(msg);
  }

  Future<String> appIn(
    Map payload, {
    encrypted = true,
    closeAfter = false,
    void Function(dynamic)? onResponse,
  }) async {
    Map<String, dynamic> p = {};

    if (config!.version == null) {
      payload = {
        "instance_id": config!.instanceID,
        "run_id": config!.runID,
        "app_id": config!.appID,
        ...payload,
      };
    } else {
      Map<String, dynamic> meta = {
        "instance_id": config!.instanceID,
        "run_id": config!.runID,
        "app_id": config!.appID,
        "version": config!.version?.toString(),
      };

      p["meta"] = meta;
    }

    p["payload"] = payload;

    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: config!.appID,
      robotID: config!.robotID,
      userID: config!.userID,
      msgType: QMT_AppIn,
      //
      payload: utf8.encode(jsonEncode(p)) as Uint8List?,
      // do not remove the unnecessary cast since encode was returning
      // List<int> without it and caused an exception.
      encrypted: encrypted,
      version: config!.version,
    );

    await sendAppMsg(msg);

    if (onResponse != null) {
      subscribe(msg.id!, onResponse);
    }

    if (closeAfter == true) {
      _channel.sink.close();
    }

    return msg.id!;
  }

  void subscribe(String id, void Function(dynamic) f) {
    StreamController<dynamic> respStream =
        StreamController<dynamic>.broadcast();
    _streams[id] = respStream;
    respStream.stream.first.then((value) {
      f(value);
      respStream.close();
      _streams.remove(id);
    });
  }

  Future<String> _getUploadLink(String file) async {
    String url = "${this._apiURL}/v1/apps.upload.get";
    var resp = await _request(
      method: "GET",
      url: url,
      queryParams: {
        "app_id": config!.appID ?? "",
        "run_id": config!.runID ?? "",
        "instance_id": config!.instanceID ?? "",
        "file": file,
      },
    );

    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }

    return Uri.decodeFull(resp["url"]);
  }

  Future<InstanceUploadLinkResponse?> _getInstanceUploadLink(
    String file,
    bool public,
  ) async {
    String url = "${this._apiURL}/v1/apps.instance.files.upload";

    var resp = await _request(
      method: "GET",
      url: url,
      queryParams: {
        "app_id": config!.appID ?? "",
        "run_id": config!.runID ?? "",
        "instance_id": config!.instanceID ?? "",
        "file": file,
        "public": public.toString(),
      },
    );

    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }

    return InstanceUploadLinkResponse.fromJson(resp);
  }

  // uploadFile uploads a file to s3 bucket and returns a public url
  // if public parameters is true or null otherwise.
  Future<String?> uploadFile(
    String file,
    Uint8List data, {
    bool? public,
  }) async {
    String? url, publicUrl;
    Map<String, String> headers = {};

    if (public == null || public == false) {
      url = await _getUploadLink(file);
    } else {
      InstanceUploadLinkResponse? resp =
          await _getInstanceUploadLink(file, public);

      url = resp?.url;
      publicUrl = resp?.publicUrl;
    }

    if (url == null) {
      throw Exception("url not found");
    }

    if (public != true) {
      data = await this.encrypt!(data, aesKey.bytes);
    }

    if (public == true) {
      headers["x-amz-acl"] = "public-read";
    }

    Uri uri = Uri.parse(url);
    await html.HttpRequest.request(
      uri.toString(),
      method: "PUT",
      mimeType: "application/octet-stream",
      responseType: "json",
      sendData: data,
      requestHeaders: headers,
    );

    return publicUrl;
  }

  Future<String?> getFileURL(String file) async {
    String url = "${this._apiURL}/v1/apps.file.get";
    var resp = await _request(
      method: "GET",
      url: url,
      queryParams: {
        "app_id": config!.appID ?? "",
        "file": file,
        "run_id": config!.runID ?? "",
        "instance_id": config!.instanceID ?? "",
      },
    );

    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }
    return resp["url"];
  }

  String _getAPIUrl() {
    String ws = this.config!.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0];
    return domain == "robomotion.io" ? "$protocol//api.$domain" : "$ws/api";
  }

  String _getAppUrl() {
    String ws = this.config!.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0] == "https:" ? "wss:" : "ws:";
    return domain == "robomotion.io"
        ? "$protocol//mq.$domain/app"
        : "$protocol//${parts[0].split("//")[1]}.$domain/mq/app";
  }

  Future<Map?> getUser() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/users.get',
    );

    return resp['user'];
  }

  Future<Map<String, Vault>> getVaults() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/vaults.list',
    );

    Map<String, Vault> vaults = {};
    for (var v in resp['vaults']) {
      vaults[v['id']] = Vault(vault: v);
    }

    return vaults;
  }

  Future<Map<String, dynamic>> getVaultItems(String vaultID) async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/vaults.items.list',
      queryParams: {'vault_id': vaultID},
    );

    Map<String, dynamic> items = {};
    for (var i in resp['items']) {
      items[i['id']] = i;
    }

    return items;
  }

  Future<Map<String, dynamic>> getConnectedRobots() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/workspace.robots.connected.list',
    );

    Map<String, dynamic> robots = {};
    for (var r in resp['robots']) {
      robots[r['id']] = r;
    }

    return robots;
  }

  Future<String> getPrivateKey(Map user, String password) async {
    final encPrivateKey = user['enc_private_key'];
    final salt = user['salt'];
    final iv = user['iv'];
    final username = user['email'];

    final algorithm = Hkdf(hmac: Hmac(Sha256()), outputLength: 32);
    final derivedSalt = await algorithm.deriveKey(
      secretKey: SecretKey(utf8.encode(username)),
      nonce: hex.decode(salt),
    );

    final pbkdf2 = Pbkdf2(
      macAlgorithm: Hmac.sha256(),
      iterations: 100000,
      bits: 256,
    );

    final mk = await pbkdf2.deriveKey(
      secretKey: SecretKey(utf8.encode(password)),
      nonce: utf8.encode(hex.encode(await derivedSalt.extractBytes())),
    );

    final _encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(await mk.extractBytes()),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final pkPem = _encrypter.decryptBytes(
      enc.Encrypted(Uint8List.fromList(hex.decode(encPrivateKey))),
      iv: enc.IV(Uint8List.fromList(hex.decode(iv))),
    );

    return utf8.decode(pkPem);
  }

  Future<void> removeVault(String vaultID) async {
    await apiRequest(
      method: 'DELETE',
      endpoint: '/v1/vaults.delete',
      body: {
        'vault_id': vaultID,
      },
    );
  }

  Future<Vault?> createVault(
    Map user,
    String name,
    String description, {
    List<int>? secretKey,
  }) async {
    var random = Random.secure();
    var vaultKey = List<int>.generate(32, (i) => random.nextInt(256));

    if (secretKey == null) {
      secretKey = List<int>.generate(32, (i) => random.nextInt(256))
          .toSet()
          .toList()
          .sublist(0, 16);

      secretKey = [...secretKey, ...secretKey];
    }

    if (secretKey.length != 32) {
      throw Exception("Invalid secret key.");
    }

    var iv = List<int>.generate(16, (i) => random.nextInt(256));
    var hexVaultKey = hex.encode(vaultKey);

    Uint8List vk = Uint8List.fromList(vaultKey);
    Uint8List sk = Uint8List.fromList(secretKey);
    var xored = vk.xor(sk);
    var xoredKey = xored.toList();

    final encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(xoredKey),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final verifier = encrypter
        .encrypt(
          hexVaultKey,
          iv: enc.IV(
            Uint8List.fromList(iv),
          ),
        )
        .base64;

    final pubPem = user['public_key'];
    final encVaultKey =
        await rsa.RSA.encryptOAEPBytes(vk, "", rsa.Hash.SHA256, pubPem);
    final encSecretKey =
        await rsa.RSA.encryptOAEPBytes(sk, "", rsa.Hash.SHA256, pubPem);

    final encVaultKeyb64 = base64.encode(encVaultKey);
    //final secretKeyb64 = base64.encode(secretKey);
    final encSecretKeyb64 = base64.encode(encSecretKey);

    final vaultIV = hex.encode(iv);

    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.create',
      body: {
        'name': name,
        'description': description,
        'enc_vault_key': encVaultKeyb64,
        'verifier': verifier,
        'iv': vaultIV,
      },
    );

    return Vault(
      vault: resp['vault'],
      secretKey: secretKey,
      vaultKey: vaultKey,
      encSecretKey: encSecretKeyb64,
      encVaultKey: encVaultKeyb64,
    );
  }

  Future<String> _encryptVaultItem(
    List<int> secretKey,
    List<int> vaultKey,
    List<int> iv,
    String data,
  ) async {
    var xored = Uint8List.fromList(secretKey).xor(
      Uint8List.fromList(vaultKey),
    );

    final encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(xored),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final encItem = encrypter.encrypt(
      data,
      iv: enc.IV(
        Uint8List.fromList(iv),
      ),
    );

    return hex.encode(encItem.bytes.toList());
  }

  Future<String> _decryptVaultItem(
    List<int> secretKey,
    List<int> vaultKey,
    List<int> iv,
    String data,
  ) async {
    var xored = Uint8List.fromList(secretKey).xor(
      Uint8List.fromList(vaultKey),
    );

    final encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(xored),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final decItem = encrypter.decrypt(
      enc.Encrypted(Uint8List.fromList(hex.decode(data))),
      iv: enc.IV(
        Uint8List.fromList(iv),
      ),
    );

    return decItem;
  }

  Future<Map?> createVaultItem(
    String vaultID,
    List<int> secretKey,
    List<int> vaultKey,
    String name,
    int category,
    String data,
  ) async {
    final random = Random.secure();
    var iv = List<int>.generate(16, (i) => random.nextInt(256));

    final encVaultItem = await _encryptVaultItem(
      secretKey,
      vaultKey,
      iv,
      data,
    );

    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.items.create',
      body: {
        'vault_id': vaultID,
        'name': name,
        'category': category,
        'data': encVaultItem,
        'iv': hex.encode(iv),
      },
    );

    return resp['item'];
  }

  Future<void> updateVaultItem(
    String vaultID,
    String itemID,
    List<int> secretKey,
    List<int> vaultKey,
    String iv, {
    String name = "",
    String data = "",
  }) async {
    List<int> ivHex = hex.decode(iv);

    final encVaultItem = await _encryptVaultItem(
      secretKey,
      vaultKey,
      ivHex,
      data,
    );

    await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.items.update',
      body: {
        'vault_id': vaultID,
        'item_id': itemID,
        'name': name,
        'data': encVaultItem,
        'iv': iv,
      },
    );
  }

  Future<String> getVaultItem(
    String vaultID,
    String itemID,
    List<int> secretKey,
    List<int> vaultKey,
  ) async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/vaults.items.get',
      queryParams: {
        "vault_id": vaultID,
        "item_id": itemID,
      },
    );

    var item = resp["item"];
    String data = resp["data"];

    return await _decryptVaultItem(
      secretKey,
      vaultKey,
      hex.decode(item?["iv"]),
      data,
    );
  }

  Future<void> injectVault(
    String vaultID,
    String robotID,
    String encSecretKey,
    String encVaultKey,
  ) async {
    await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.secret.inject',
      body: {
        'vault_id': vaultID,
        'robot_id': robotID,
        'enc_secret': encSecretKey,
        'enc_key': encVaultKey,
      },
    );
  }

  Future<String?> askPassword(BuildContext context) async {
    TextEditingController _controller = TextEditingController();
    int? resp = await showDialog<int>(
      context: context,
      builder: (BuildContext context) => AlertDialog(
        title: const Text('Welcome Back!'),
        content: Column(
          mainAxisAlignment: MainAxisAlignment.start,
          mainAxisSize: MainAxisSize.min,
          children: [
            Row(
              children: [
                Text(
                  'Password',
                  style: GoogleFonts.openSans(
                    fontSize: 14,
                    fontWeight: FontWeight.w600,
                  ),
                ),
              ],
            ),
            SizedBox(
              width: 240,
              child: TextFormField(
                obscureText: true,
                enableSuggestions: false,
                autocorrect: false,
                style: GoogleFonts.openSans(
                  fontSize: 14,
                  fontWeight: FontWeight.w500,
                ),
                decoration: const InputDecoration(
                  icon: Icon(
                    Icons.password,
                  ),
                  enabledBorder: UnderlineInputBorder(
                    borderSide: BorderSide(),
                  ),
                  hintText: 'Your Password',
                ),
                controller: _controller,
              ),
            ),
          ],
        ),
        actions: [
          ElevatedButton(
            style: ButtonStyle(
              padding: MaterialStateProperty.all(
                const EdgeInsets.all(16),
              ),
              backgroundColor: MaterialStateProperty.all(Colors.blue),
              shape: MaterialStateProperty.all(
                RoundedRectangleBorder(
                  borderRadius: BorderRadius.circular(8),
                ),
              ),
            ),
            onPressed: () {
              Navigator.pop(context, 0);
            },
            child: Text(
              'Next',
              style: GoogleFonts.openSans(
                fontSize: 14,
                fontWeight: FontWeight.w600,
                color: Colors.white,
              ),
            ),
          ),
        ],
      ),
    );

    return resp == null ? null : _controller.text;
  }

  Future<String?> askVaultSecret(BuildContext context, Vault vault) async {
    TextEditingController _controller = TextEditingController();
    int? resp = await showDialog<int>(
      context: context,
      builder: (BuildContext context) => AlertDialog(
        title: const Text('Unlock Vault'),
        content: Column(
          mainAxisAlignment: MainAxisAlignment.start,
          mainAxisSize: MainAxisSize.min,
          children: [
            Row(
              children: [
                Text(
                  'Vault Secret Key',
                  style: GoogleFonts.openSans(
                    fontSize: 14,
                    fontWeight: FontWeight.w600,
                  ),
                ),
              ],
            ),
            SizedBox(
              width: 240,
              child: TextFormField(
                obscureText: true,
                enableSuggestions: false,
                autocorrect: false,
                style: GoogleFonts.openSans(
                  fontSize: 14,
                  fontWeight: FontWeight.w500,
                ),
                decoration: const InputDecoration(
                  icon: Icon(
                    Icons.password,
                  ),
                  enabledBorder: UnderlineInputBorder(
                    borderSide: BorderSide(),
                  ),
                  hintText: 'Type the secret key here',
                ),
                controller: _controller,
              ),
            ),
          ],
        ),
        actions: [
          ElevatedButton(
            style: ButtonStyle(
              padding: MaterialStateProperty.all(
                const EdgeInsets.all(16),
              ),
              backgroundColor: MaterialStateProperty.all(Colors.blue),
              shape: MaterialStateProperty.all(
                RoundedRectangleBorder(
                  borderRadius: BorderRadius.circular(8),
                ),
              ),
            ),
            onPressed: () {
              Navigator.pop(context, 0);
            },
            child: Text(
              'Unlock',
              style: GoogleFonts.openSans(
                fontSize: 14,
                fontWeight: FontWeight.w600,
                color: Colors.white,
              ),
            ),
          ),
        ],
      ),
    );

    return resp == null ? null : _controller.text;
  }

  Future<void> verifySecret(
    Vault vault,
    Map user,
    String password,
  ) async {
    final vaultKey = await vault.getVaultKey(user, password);
    final secretKey = await vault.getSecretKey(user, password);

    Uint8List vk = Uint8List.fromList(vaultKey);
    Uint8List sk = Uint8List.fromList(secretKey);

    final xored = vk.xor(sk);

    final String iv = vault.vault["iv"];
    final String verifier = vault.vault["verifier"];

    final _encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(xored),
        mode: enc.AESMode.cbc,
      ),
    );

    var decVerifier = _encrypter.decryptBytes(
      enc.Encrypted(Uint8List.fromList(base64.decode(verifier))),
      iv: enc.IV(Uint8List.fromList(hex.decode(iv))),
    );

    decVerifier = hex.decode(utf8.decode(decVerifier));

    if (!listEquals(decVerifier, vaultKey)) {
      throw Exception("Vault secret key not verified.");
    }
  }
}

class Config {
  String? userID = "";
  final String? workspace;
  String? appID;
  String? robotID;
  String? runID;
  String? flowID;
  String? versionID;
  String? instanceID;
  String? token;
  late Map _profile;
  Version? version;

  Config({this.workspace, this.userID, this.token}) {
    var cookie = _getCookie("profile");
    if (cookie != null) {
      this._profile =
          jsonDecode(Uri.decodeFull(cookie)) ?? "{}" as Map<dynamic, dynamic>;
      this.userID = this._profile["id"] ?? "";
    }
  }
}

String? _getCookie(String cname) {
  Map _cookies = {};
  html.document.cookie
      ?.split(";")
      .map(
        (c) => c.split("="),
      )
      .where((c) => c.length > 1)
      .forEach(
        (c) => _cookies[c[0].trim()] = c[1],
      );

  return _cookies[cname];
}
